use crate::config::node::NodeType;
use log::{debug, error, info, trace};
use std::{env, io};


use std::cell::RefCell;
use std::future::Future;
use std::net::SocketAddr;
use std::str::FromStr;
use tokio::net::UdpSocket;

pub mod config;
pub mod connection;

thread_local! {
    static CONFIG: RefCell<config::Config> = RefCell::new(config::Config::from_path(
        env::var("DOMO_CONFIG_PATH")
            .unwrap_or("config.toml".into())
    ).unwrap_or({ info!("Using default config..."); config::Config::default() }))
}

fn main() -> io::Result<()> {
    setup_logging().expect("could not setup logging");

    debug!("Building tokio runtime...");

    let runtime = tokio::runtime::Builder::new_multi_thread()
        .enable_all()
        .build()
        .unwrap();

    debug!("Built tokio runtime with all features...");

    info!(
        "Staring as {} node!",
        CONFIG.with_borrow(|c| c.node.node_type.clone())
    );


    match CONFIG.with_borrow(|c| c.node.node_type.clone()) {
        NodeType::Master { bind, identifier } => runtime.block_on(async {
            let sock = UdpSocket::bind(bind).await.unwrap();
            info!("bound to: {}", sock.local_addr().unwrap());
            loop {
                let mut buf = vec![0; domo_proto::packet::FULL_PACKET_SIZE];
                match sock.recv_from(&mut buf).await {
                    Ok((size, addr)) => {
                        trace!("recv[{addr} -> {size}]: {:?}", &buf[..size]);
                    },
                    Err(e) => {
                    }
                }
            }
        }),
        NodeType::Relay { bind, master_address } => runtime.block_on(async {
        }),
        NodeType::Subnet { master_address: _ } => {
            runtime.block_on(async {})
        }
    }

    Ok(())
}

fn setup_logging() -> Result<(), fern::InitError> {
    fern::Dispatch::new()
        .format(|out, message, record| match record.metadata().target() {
            _ => out.finish(format_args!(
                "[{} - {}] <{}> {}: {}",
                humantime::format_rfc3339(std::time::SystemTime::now()),
                record.level(),
                record.target(),
                record.file().unwrap_or("??"),
                message
            )),
        })
        .chain({
            fern::Dispatch::new()
                .level(CONFIG.with_borrow(|c| c.log.stdout_level.clone()).into())
                .chain(io::stdout())
        })
        .chain({
            fern::Dispatch::new()
                .level(CONFIG.with_borrow(|c| c.log.stdout_level.clone()).into())
                .chain(fern::log_file("./node.log")?)
        })
        .apply()?;

    Ok(())
}
